//INIZIO CODICE
//
//
//Inizio listato
//prodotto da calcola (antrec@geocities.com)//
//ver.5_3_3
//
//ATTENZIONE:
//
//CC BY-NC-SA 4.0
//Uso non commerciale, modificabile, con obbligo di condivisione-
//http://creativecommons.org/licenses/by-nc-sa/4.0/deed.it
//
/* Cosa fa:
Legge la tensione delle batterie su A0 e la usa per la commutazione enel/batterie con inverter UPS, al raggiungimento della tensione minima attende per un numero di secondi 
pari alla variabile config.pausa e poi controlla se le batterie sono ancora alla tensione minima, se vero commuta su enel altrimenti resta a batterie; dopo la commutazione
può attivare un relè per comandare un alimentatore di ricarica ausiliario.  Sui pin A1 e A2 legge la
corrente del pannello e delle batterie, calcola la produzione ed il consumo istantanei, orari, giornalieri, memorizza i dati della produzione/consumo del giorno prima.
Permette di impostare le variabili preferenziali da momitor seriale senza dover ricaricare il listato, su monitor seriale trasmette anche alcune 
informazioni sul funzionamento. Replica alcune informazioni su monitor LCD
Con la variabile config.stato_carica e tramite il pin 9 è possibile comandare un caricabatterie ausiliario per la ricarica d'emergenza,  
la soglia minima d'intervento è settabile da monitor seriale, la soglia di fine ricarica è settabile e la ripresa della ricarica, se le batterie si scaricano nuovamente,
si setta con il parametro isteresi.
Salva i dati su SD e l'intervallo è configurabile, il pin CS è collegato al pin digitale 17 (A3).
//
*/
// LIBRERIE
#include "OneWire.h"
#include "Wire.h"
#include "avr/eeprom.h"
#include "RTClib.h"
#include <avr/io.h>
#include <avr/wdt.h>
#include <LiquidCrystal_I2C.h>
#include <SPI.h>
#include <SD.h>

//IMPOSTAZIONI
#define CONFIG_EEPROM_ADDR ((byte*) 0x10)
RTC_DS1307 RTC;
LiquidCrystal_I2C lcd(0x27,20,4);
OneWire  ds(3); //Setto come pin per la onewire (pwm!!!) la usa la sonda termica da saldare sul modulo RTC


//VARIABILI MEMORIZZATE
struct Config {
    float tensione_bassa;
    float tensione_alta;
    byte pausa;
    float coefficiente_conversione;
    float stop_ricarica;
    float ist;
    byte tempo_salva;
    byte stato_carica;
    byte valid; //lasciare alla fine!
} config;

//IMPOSTAZIONI VARIABILI IN MEMORIA (configurazione caricata all'avvio se eprom vuota)
static void loadConfig() {
    for (byte i = 0; i < sizeof config; ++i)
        ((byte*) &config)[i] = eeprom_read_byte(CONFIG_EEPROM_ADDR + i);
        if (config.valid != 253) {
        config.valid = 253;
//
////////////////////////////////
//INIZIO PARTE MODIFICABILE
///////////////////////////////
        
        config.coefficiente_conversione = 0.0552; //arrotondare alla 4.a cifra decimale
        config.tensione_bassa = 11.9;    //valori consigliati: 11.9 (sistema 12V); 23.8 (sistema 24V)
        config.tensione_alta = 13.4;     //valori consigliati: 13.7 (sistema 12V); 27.4 (sistema 24V)
        config.pausa = 5;                //consigliato 5, ritardo in secondi per compensare brevi abbassamenti di tensione, ATTENZIONE solo valori interi
        config.stop_ricarica = 12.4;     //valori consigliati: 12.4 (sistema 12V); 24.8 (sistema 24V)
        config.ist = 0.5;                //isteresi ricarica
        config.tempo_salva = 3;  //intervallo tempo in MINUTI per campionamento dati su SD - ATTENZIONE, SOLO VALORI INTERI
        config.stato_carica = 1;  // valori: 1=si;  0=no
        
///////////////////////////////
//FINE PARTE MODIFICABILE
///////////////////////////////
//        
        }
        }
        
//SALVATAGGIO IN MEMORIA DELLE VARIABILI
    static void saveConfig() {
    for (byte i = 0; i < sizeof config; ++i)
        eeprom_write_byte(CONFIG_EEPROM_ADDR + i, ((unsigned char*) &config)[i]);        
}

//DEFINIZIONI
#define curr_pan (A1) //sensore corrente pannello
#define curr_batt (A2) //sensore corrente batterie
#define R_batt 4 //led verde + relè commutazione. Indicatore stato batterie
#define R_enel 5 //led rosso. Indicatore stato rete.
#define RICARICA 9 //Ricarica batterie
#define sen_tens (A0) //Sensore tensione collegato al pin analogico 0
#define numero_letture 5 //Il numero indica i valori tensione da leggere
#define led 8 //Led stato attivo SD
#define chipSelect  17 // pin da collegare al CS del lettore SD in alternativa un altro pin digitale libero es il 10

//DEFINIZIONE COSTANTI
#define bassatensione (config.tensione_bassa/config.coefficiente_conversione) // Valore numerico della tensione stop batt ritorno rete
#define altatensione (config.tensione_alta/config.coefficiente_conversione) // Valore numerico tensione stop rete passo batt.
#define media1 (bassatensione + altatensione)/2 // Media tensione per avvio numero
#define media_avvio media1*config.coefficiente_conversione // media1 in Volt

//VARIABILI
        
int rit;
int c;
unsigned long a;
unsigned long b;
unsigned long salv;
byte stato_salv = 0;
byte min_avvio;//tempo avvio
byte min_avvio_ora;//tempo avvio conto secondi per media oraria
byte min_attuale;//tempo per il campionamento
byte blocco = 0;
byte blocco_ora = 0;
byte blocco_gg = 0;
byte giorno;
byte sec_attuale;
byte ora_attuale;
int n_letture = 0;//numero letture istantanee il contatore si azzera ogni ora
int n_letture_cont = 0;// il contatore si azzera in caso di riavvio
int readings[numero_letture]; //Letture da ingresso analogico
int index = 0; //Indice della lettura corrente
int total = 0; //Totale letture
int media_tens = 0; //Media tensione indice analogico
float media_tens_volt; //media tensione espressa in Volt
byte enel_s = 0;
byte stato = 0;
byte carica = 0;
float correntebatt = 0; //media corrente singola lettura
float correntepan = 0; //media corrente singola lettura
float correntepansom = 0;// somma corrente pannello un campionamento al minuto
float correntebattsom = 0;//somma corrente batterie un campionamento al minuto
float corr_pro_pan = 0; //corrente prodotta pannello giorno in corso
float corr_pro_bat = 0; //bilancio corrente in batteria giorno in corso
float corr_ba_med = 0; //media corrente oraria batteria
float corr_pa_med = 0; //media corrente oraria pannello
float corr_prod_p_i = 0; //corrente prodotta pannello giorno precedente
float corr_prod_b_i = 0; //bilancio corrente batteria giorno precedente
char Nome_data[10];


void setup(){
  
Serial.begin(9600);               //Apro la comunicazion seriale
lcd.init();                      // apro LCD
lcd.backlight();
lcd.clear();
Wire.begin();
DateTime now = RTC.now();
min_avvio = (now.minute());
min_avvio_ora = min_avvio;
  
pinMode(led, OUTPUT);
pinMode(R_batt, OUTPUT);
pinMode(R_enel, OUTPUT);
pinMode(RICARICA, OUTPUT);
pinMode(sen_tens, INPUT);
pinMode(curr_pan, INPUT);
pinMode(curr_batt, INPUT);

loadConfig();

//INIZIALIZZO SD E CREO FILE

Serial.println("");
Serial.println("v: 5_3_3");
Serial.println("Accesso SD card...");
pinMode(chipSelect, OUTPUT);

//FISSO IL GIORNO E RIMANDO ALLA ROUTINE INTESTAZIONE DEL FILE DA SALVARE SU SD
giorno = now.day(); intestazione();

   
//SISTEMA RTC SE NON HA ORA MEMORIZZATA
if (! RTC.isrunning()) {
    RTC.adjust(DateTime(__DATE__, __TIME__));
   }

//abilito il watchdog e imposto come tempo di reset 8 secondi
wdt_enable(WDTO_8S);  


//SETTO STATO RELE' ALL'AVVIO
{if (analogRead(sen_tens) > media1){batterie();} 
else{enel();}}

//LIBERO MEMORIA

F(" Tensione batteria (media 5 letture): ");
F(" Tensione batteria intermedio tra min-max: ");
F(" Passa a enel quando minore di: ");
F(" Passa a batterie quando maggiore di: ");
F(" Tensione limite ricarica: ");
F(" Coefficiente di conversione: ");
F(" Ritardo commutazione (consigliato 5): ");
F(" Corrente istantanea pannello (A): ");
F(" Corrente media pannello (Ah): ");
F(" Corrente totale pannello (Ah): ");
F(" Corrente istantanea batterie (A): ");
F(" Corrente media batterie (A): ");
F(" Corrente totale batterie (A): ");
F(" Produzione totale pannello ieri (Ah): ");
F(" Produzione totale batterie ieri (Ah): ");
F("Tensione commutazione enel: ");
F("Tensione commutazione batterie: ");
F("Tensione fine ricarica: ");
F("Isteresi ricarica: ");
F("Coefficiente di conversione: ");
F("Secondi ritardo commutazione: ");
F("Errore, ripeti");
F("errore apertura file dati  ");
F("-----------");
F("=============================================");
F("********************");
F("    Data              ");
F("Corr. pannello         ");
F("Corr. batterie       ");
F("Tens. batterie             ");
F(" Minuti al salvataggio dati: ");
F("Stato: batterie");
F(" Stato: enel");
F(" Stato: enel + ricarica");
F("Minuti al campionamento su SD: ");
F("Attivo la funzione ricarica (1=si; 0=no): ");

}

void loop(){
  

tensione();
sensore_corrente();
ricarica();
stampa();

//TEMPO SALVATAGGIO SU SD

DateTime now = RTC.now();

//Controllo se il valore della variabile giorno è diverso dal giorno di calendario se vero creo un nuovo file
if (giorno  != now.day()){giorno = now.day(); intestazione();}

// salvo i dati una volta raggiunto il valore impostato nella variabile tempo_salva
if (millis() > (salv + (config.tempo_salva*60000)) && (stato_salv == 1 )){stato_salv = 0; dati();}//60000 SALVA OGNI MINUTO, VARIARE SE OCCORRE
if (stato_salv == 0){salv = millis(); stato_salv = 1;}

//FINE SALVATAGGIO

if (Serial.available() > 0){ immissione();} //leggo seriale se arrivano dati da pc vado su routine immissione

//Calcolo lo stato del relè di commutazione in base alla tensione letta
if ((media_tens_volt < config.tensione_bassa) && (enel_s !=1) && (stato !=1))
{ritardo();}
  if (stato == 1){b = millis(); if ((b-a > rit) && (media_tens_volt < config.tensione_bassa)){enel();}}
  if(media_tens_volt > config.tensione_alta){batterie();}
  if (enel_s == 0) {Serial.println(F("Stato: batterie"));Serial.println("");}
  else if ((enel_s == 1) && (carica == 0)){Serial.println(F(" Stato: enel"));Serial.println("");}
  else if ((enel_s == 1) && (carica == 1)){Serial.println(F(" Stato: enel + ricarica"));Serial.println("");}

//Alla fine del ciclo loop rimando alla routine riavvio del contatore del watchdog
riavvio();
}

void stampa(){
  
    ///////// SU SERIALE

Serial.println(F("============================================="));
Serial.print(F(" Tensione batteria intermedio tra min-max: "));
Serial.println(media_avvio);
Serial.println();
Serial.print(F(" Tensione batteria (media 5 letture): "));
Serial.println(media_tens_volt); //Serial.println(media_tens);
Serial.print(F(" Passa a enel quando minore di: "));
Serial.println(config.tensione_bassa);
Serial.print(F(" Passa a batterie quando maggiore di: "));
Serial.println(config.tensione_alta);
//Serial.println("-----------");
Serial.print(F(" Tensione limite ricarica: "));
Serial.println(config.stop_ricarica);
Serial.print(" Isteresi ricarica: ");
Serial.println(config.ist);
Serial.print(F(" Coefficiente di conversione: "));
Serial.println(config.coefficiente_conversione,4);
Serial.println(F("-----------"));
Serial.print(F(" Ritardo commutazione (consigliato 5): "));
Serial.println(config.pausa);
Serial.println(F("-----------"));
Serial.print(F(" Corrente istantanea pannello (A): "));
Serial.println(correntepan);
Serial.print(F(" Corrente media pannello (Ah): "));
Serial.println(corr_pa_med);
Serial.print(F(" Corrente totale pannello (Ah): "));
Serial.println(corr_pro_pan);
Serial.println(F("-----------"));
Serial.print(F(" Corrente istantanea batterie (A): "));
Serial.println(correntebatt);
Serial.print(F(" Corrente media batterie (A): "));
Serial.println(corr_ba_med);
Serial.print(F(" Corrente totale batterie (A): "));
Serial.println(corr_pro_bat);
Serial.println(F("-----------"));
Serial.print(F(" Produzione totale pannello ieri (Ah): "));
Serial.println(corr_prod_p_i);
Serial.print(F(" Produzione totale batterie ieri (Ah): "));
Serial.println(corr_prod_b_i);
Serial.println(F("-----------"));
Serial.print(F(" Minuti al salvataggio dati: "));
Serial.println((config.tempo_salva)-((millis()-salv)/60000));
Serial.println("v: 5_3_3");



/////// SU LCD---LCD---
lcd.home();
lcd.noBlink();
lcd.setCursor(0,1);
lcd.print("T:"); lcd.print(media_tens_volt,1);
lcd.setCursor(7,1);lcd.print("S:");
if (enel_s == 0 ){lcd.print("Batt  ");}
if ((enel_s == 1) && (carica != 1)){ lcd.print("Enel  ");}
if ((enel_s == 1) && (carica == 1)){ lcd.print("EnelR ");}
lcd.print((config.tempo_salva)-((millis()-salv)/60000));
lcd.setCursor(0,0); lcd.print("                ");
lcd.setCursor(0,0); lcd.print("P:"); lcd.print(correntepan,1);
lcd.setCursor(7,0);
lcd.print(" B:"); lcd.print(correntebatt,1); 
}

//// CREO LA RIGA DI INTESTAZIONE DI UN NUOVO FILE

void intestazione(){
DateTime now = RTC.now();
  
 String Nome_data = (String(now.day()) + "_" + String(now.month()) + ".txt");

(SD.begin(chipSelect ));
File File_dati = SD.open(Nome_data.c_str(), FILE_WRITE);

if (File_dati) {File_dati.print(now.day());File_dati.print("/ ");File_dati.print(now.month()); File_dati.print("/ "); File_dati.println(now.year());
  File_dati.print(F(" ORA          "));File_dati.print(F("Corr. pannello         ")) ; File_dati.print(F("Corr. batterie       ")) ;
File_dati.print(F("Tens. batterie        ")) ; File_dati.println("   Stato  ");Serial.println("card pronta.");
Serial.println(F("********************"));Serial.println("Dati salvati.  "); Serial.println(F("********************"));digitalWrite(led, HIGH);File_dati.close();SD.end();
digitalWrite(led, HIGH);
}else{Serial.println("SD rotta o non presente");digitalWrite(led, LOW);lcd.print(" Inserire SD");while(1);
}
  
}

//// SALVO DATI SD

void dati(){

DateTime now = RTC.now();

String Nome_data = (String(now.day()) + "_" + String(now.month()) + ".txt");
//Serial.println(Nome_data);

if (SD.begin(chipSelect));
File File_dati = SD.open(Nome_data.c_str(), FILE_WRITE);
if (File_dati){ File_dati.print(ora_attuale);File_dati.print(": ");File_dati.print(min_attuale);
   File_dati.print(";            ");File_dati.print(correntepan,1); File_dati.print(";               ");File_dati.print(correntebatt,1);File_dati.print(";              ");
   File_dati.print(media_tens_volt,1);File_dati.print(";              ");
if (enel_s == 0 ){File_dati.println("    batteria ");}
if ((enel_s == 1) && (carica != 1)){ File_dati.println("      Enel  ");}
if ((enel_s == 1) && (carica == 1)){ File_dati.println("      Enel+Ricarica");}
Serial.println(F("********************"));Serial.println("Dati salvati.  "); Serial.println(F("********************"));digitalWrite(led, HIGH);File_dati.close();SD.end();
}
else{Serial.println(F("********************"));Serial.println(F("errore apertura file dati  "));Serial.println(F("********************"));digitalWrite(led, LOW);File_dati.close();
SD.end();}
}

///// MISURO LA TENSIONE DELLE BATTERIE

void tensione(){
  
total=0;  
for (int A = 0; A < numero_letture; A++){
total=total + analogRead(sen_tens); }

media_tens = total / numero_letture; //Calcola la media

media_tens_volt = media_tens * config.coefficiente_conversione;

}

//////// MISURO LA CORRENTE GENERATA DAL PANNELLO E QUELLA I/O DELLE BATTERIE

void sensore_corrente(){
 
  for(int s = 0; s < 40; s++) {
  correntepan = correntepan + ((0.245 * analogRead(curr_pan)) - 124.87); //pannello sensore ACS758-100B (0,245 125  0,01718)
  correntebatt = correntebatt + ((0.49 * analogRead(curr_batt)) - 250.16);  //batterie sensore ACS758-200B  (0,49 250)  
  }
 
 correntepan = correntepan/40;
 correntebatt = correntebatt/40;
 
 DateTime now = RTC.now();
 min_attuale = (now.minute());
 sec_attuale = (now.second());
 ora_attuale = (now.hour());
 
 if (sec_attuale == 0 && blocco == 0){n_letture = n_letture + 1;  n_letture_cont = n_letture_cont + 1;
    correntebattsom = (correntebattsom + correntebatt);  
     correntepansom = (correntepansom + correntepan);  
     corr_ba_med = correntebattsom / n_letture;  
     corr_pa_med = correntepansom / n_letture;
     blocco = 1;}
 if (sec_attuale == 3){blocco = 0;}

 if (min_attuale == 0 && blocco_ora == 0){
     corr_pro_bat = corr_pro_bat + corr_ba_med;
     corr_pro_pan = corr_pro_pan + corr_pa_med;
     correntepansom = 0; correntebattsom = 0; n_letture = 0; blocco_ora = 1;}
 if (min_attuale == 1){blocco_ora = 0;}
 
 if  (ora_attuale == 0 && blocco_gg == 0){corr_prod_b_i = corr_pro_bat; corr_prod_p_i = corr_pro_pan; 
 corr_ba_med = 0; corr_pa_med = 0; correntepansom = 0;correntebattsom = 0;char strnome[6] = {now.day(),'_',now.month(), '\ 0'} ;blocco_gg = 1;}
 if  (ora_attuale == 1){blocco_gg = 0;}
 
}

//// ROUTINE CHE ATTUA LA COMMUTAZIONE SU ENEL

void enel() {

digitalWrite(R_enel, HIGH);
delay(200);
digitalWrite(R_batt, LOW);
a=b;
enel_s = 1;
stato = 2 ;
}

//// ROUTINE CHE ATTUA LA COMMUTAZIONE SU BATTERIA

void batterie(){

digitalWrite(R_batt, HIGH);
delay(200);
digitalWrite(R_enel, LOW);
a=b;
enel_s = 0;
stato = 2 ;
}

//// ROUTINE PER L'ATTIVAZIONE DELLA RICARICA D'EMERGENZA

void ricarica(){
  if ((config.stop_ricarica > media_tens_volt) && (enel_s == 1) && (config.stato_carica == 1)) {(digitalWrite(RICARICA, HIGH)); carica = 1;}
  if ((enel_s == 1) && ((config.stop_ricarica -config.ist) < media_tens_volt) && (config.stato_carica == 1)){(carica = 0);(digitalWrite(RICARICA, LOW));}
  }

//// ROUTINE PER IL RITARDO NELLA COMMUTAZIONE
void ritardo(){
rit = config.pausa*1000;
a= millis();
stato = 1;
}

//// IMMISSIONE PARAMETRI CONFIGURAZIONE DA SERIALE
void immissione(){
 c= Serial.findUntil("soglie", "soglie");
 if (c == 1){svuotabuffer(); c = 0;
 loadConfig();
Serial.println(F("-----------"));Serial.println(F("-----------"));Serial.println(F("-----------"));
Serial.println(F("============================================="));
Serial.print(F("Tensione commutazione enel: "));
delay(4000); wdt_reset();
if (Serial.available() > 0){config.tensione_bassa = Serial.parseFloat(); svuotabuffer();}
wdt_reset(); Serial.println(config.tensione_bassa);
Serial.println("");

Serial.print(F("Tensione commutazione batterie: "));
delay(4000); wdt_reset();
if (Serial.available() > 0){ config.tensione_alta = Serial.parseFloat(); svuotabuffer();}
Serial.println(config.tensione_alta);
Serial.println("");

Serial.print(F("Attivo la funzione ricarica (1=si; 0=no): "));
delay(4000); wdt_reset();
if (Serial.available() > 0){ config.stato_carica = Serial.parseFloat();svuotabuffer();}
Serial.println(config.stato_carica);
Serial.println("");

Serial.print(F("Tensione fine ricarica: "));
delay(4000); wdt_reset();
if (Serial.available() > 0){ config.stop_ricarica = Serial.parseFloat();svuotabuffer();}
Serial.println(config.stop_ricarica);
Serial.println("");

Serial.print(F("Isteresi ricarica: "));
delay(4000); wdt_reset(); 
if (Serial.available() > 0){ config.ist = Serial.parseFloat(); svuotabuffer();}
Serial.println(config.ist);
Serial.println("");

Serial.print(F("Coefficiente di conversione: "));
delay(5000); wdt_reset();
if (Serial.available() > 0){ config.coefficiente_conversione = Serial.parseFloat(); svuotabuffer();}
wdt_reset(); Serial.println(config.coefficiente_conversione,4);
Serial.println("");

Serial.print(F("Secondi ritardo commutazione: "));
delay(4000); wdt_reset();
if (Serial.available() > 0){ config.pausa = Serial.parseFloat();svuotabuffer();}
wdt_reset(); Serial.println(config.pausa);
Serial.println("");

Serial.print(F("Minuti al campionamento su SD: "));
delay(4000); wdt_reset();
if (Serial.available() > 0){ config.tempo_salva = Serial.parseFloat();svuotabuffer();}
wdt_reset(); Serial.println(config.tempo_salva);
Serial.println(F("============================================="));

saveConfig();

}

else Serial.println(F("Errore, ripeti"));
delay(2000);}

void svuotabuffer() {
  while (Serial.available()) {
    byte temp = Serial.read();
  }
}

void riavvio()
{
wdt_reset(); //resetto il watchdog
//Serial.println("azzerato");
}
//Fine listato
