//Inizio listato
//prodotto da calcola (antrec.geo@yahoo.com)//
//ver.3_8_4 -Senza settaggio da monitor seriale
//          -Ritardo con millis + monitor LCD
//
//ATTENZIONE:
//
//CC BY-NC-SA 4.0
//Uso non commerciale, modificabile, con obbligo di condivisione-
//http://creativecommons.org/licenses/by-nc-sa/4.0/deed.it
//

#include "Wire.h"
#include "OneWire.h"
#include <LiquidCrystal_I2C.h>

LiquidCrystal_I2C lcd(0x27,20,4);

////////////////////////////////
//INIZIO PARTE MODIFICABILE
///////////////////////////////

float coefficiente_conversione = 0.0552; //arrotondare alla 4.a cifra decimale
float tensione_bassa = 11.5; //valori: 11.9 (sistema 12V); 23.8 (sistema 24V)
float tensione_alta = 12.2; //valori: 13.7 (sistema 12V); 27.4 (sistema 24V)
byte pausa = 8; //ritardo in secondi per compensare brevi abbassamenti di tensione

///////////////////////////////
//FINE PARTE MODIFICABILE
///////////////////////////////


int rit;
int c;
unsigned long a;
unsigned long b;
#define RELAY1 4 //led verde + relè commutazione. Indicatore stato batterie
#define RELAY2 5 //led rosso. Indicatore stato rete.
#define inputPin (A0) //Sensore tensione collegato al pin analogico 0
#define numero_letture 5 //Il numero indica i valori tensione da leggere

#define bassatensione (tensione_bassa/coefficiente_conversione) // Valore numerico della tensione stop batt ritorno rete
#define altatensione (tensione_alta/coefficiente_conversione) // Valore numerico tensione stop rete passo batt.
#define media1 (bassatensione + altatensione)/2 // Media tensione per avvio numero
#define media_avvio media1*coefficiente_conversione // media1 in Volt

int readings[numero_letture]; //Letture da ingresso analogico
int index = 0; //Indice della lettura corrente
int total = 0; //Totale letture
int media_tens = 0; //Media tensione indice analogico
float media_tens_volt; //media tensione espressa in Volt
byte enel0 = 0;
byte stato = 0;

void setup()
{
  
  Serial.begin(57600);              //Apro la comunicazion seriale
  lcd.init();                      // apro LCD
  lcd.backlight();
  Wire.begin();
  pinMode(RELAY1, OUTPUT);
  pinMode(RELAY2, OUTPUT);
  pinMode(inputPin, INPUT);

  {
    if (analogRead(inputPin) > media1){
      batterie();
    }
    else{
      enel();
    }
  }

  for (int i = 0; i < numero_letture; i++)
    readings[i] = 0; //Mette tutti 0 nell'array
}

void loop(){

  tensione();

  if ((media_tens_volt < tensione_bassa) && (enel0 != 1) && (stato != 1))
  {
    ritardo();
  }
  if (stato == 1){
    b=millis(); 
    if ((b-a > rit) && (media_tens_volt < tensione_bassa)){
      enel();
    }
  }
  if(media_tens_volt > tensione_alta){
    batterie();
  }
  if (enel0 == 0){
    Serial.println("Stato: batterie");
    lcd.setCursor(0,1);
    lcd.print("Stato: batterie");
  }
  else if (enel0 == 1){
    Serial.println("Stato: enel");
    lcd.setCursor(0,1);
    lcd.print("Stato:     enel");
  }


  Serial.print(" Voltaggio batteria intermedio tra min-max: ");
  Serial.println(media_avvio);
  Serial.println("-----------");
  Serial.print(" Voltaggio batteria (media 5 letture): ");
  Serial.println(media_tens_volt); //Serial.println(media_tens);
  Serial.print(" Passa a enel quando minore di: ");
  Serial.println(tensione_bassa);
  Serial.print(" Passa a batterie quando maggiore di: ");
  Serial.println(tensione_alta);
  Serial.println("-----------");
  Serial.print("Ritardo commutazione: ");
  Serial.println(pausa);
  Serial.print("Coefficiente di conversione: ");
  Serial.println(coefficiente_conversione,4);
  Serial.println("-----------");
  Serial.println("v: 3_8_3");
  //Serial.print("Millis iniziali; ");
  //Serial.println(a);
  //Serial.print("Millis finali: ");
  //Serial.println(b);
  Serial.println("-----------");
  Serial.println();
  lcd.setCursor(0,0);
  lcd.print("Tensione: ");
  lcd.print(media_tens_volt);
  delay(2000);
}

void tensione(){

  for (int A = 0; A < numero_letture; A++){
    total -= readings[index]; //Sottrae l'ultima lettura
    readings[index] = analogRead(inputPin); //Legge i valori provenienti dal sensore e li salva nell'array
    total += readings[index]; //Aggiunge la lettura al totale
    index = (index + 1); //Incrementa l'indice

    if (index >= numero_letture)
      index = 0; //Alla fine dell'array resetta l'indice a 0

      media_tens = total / numero_letture; //Calcola la media
  }
  media_tens_volt = media_tens*coefficiente_conversione;

}

void enel()
{
  digitalWrite(RELAY2, HIGH);
  delay(200);
  digitalWrite(RELAY1, LOW);
  a = b;
  enel0 = 1;
  stato = 2 ;
}

void batterie(){
  digitalWrite(RELAY1, HIGH);
  delay(200);
  digitalWrite(RELAY2, LOW);
  a = b;
  enel0 = 0;
  stato = 2 ;
}

void ritardo(){
  rit = pausa*1000;
  a = millis();
  stato = 1;
}
//Fine listato

